import datetime

import pytest

from tests import errors


@pytest.fixture()
def derived_error():
    ts = datetime.datetime.now()
    return errors.DerivedError(name="John",
                               surname="Brown",
                               note="...",
                               age=42,
                               updated_at=ts,
                               full_name="secret",
                               box=dict())



from unittest import mock

import pytest

from izulu import root
from tests import errors


@pytest.mark.parametrize(
    "kls",
    (errors.TemplateOnlyError, errors.AttributesOnlyError),
)
@pytest.mark.parametrize("kwargs", (dict(), dict(name="John"), dict(age=42)))
def test_forbid_missing_fields_triggered(kls, kwargs):
    with pytest.raises(TypeError):
        type("TestError",
             (kls,),
             {"__features__": root.Features.FORBID_MISSING_FIELDS}
             )(**kwargs)


@pytest.mark.parametrize(
    ("kls", "kwargs"),
    (
            (errors.RootError,
             dict(field="value")),
            (errors.TemplateOnlyError,
             dict(name="John", age=42, field="field")),
            (errors.AttributesOnlyError,
             dict(name="John", age=42, field="field")),
    ),
)
def test_forbid_undeclared_fields_triggered(kls, kwargs):
    with pytest.raises(TypeError):
        type("TestError",
             (kls,),
             {"__features__": root.Features.FORBID_UNDECLARED_FIELDS}
             )(**kwargs)


@pytest.mark.parametrize(
    ("kls", "kwargs"),
    (
            (errors.ClassVarsError, dict(name="John")),
            (errors.ClassVarsError, dict(age=0)),
            (errors.ClassVarsError, dict(blah=1.0)),
            (errors.ClassVarsError, dict(name="John", age=0)),
            (errors.ClassVarsError, dict(age=0, blah=1.0)),
            (errors.MixedError, dict(name="John", age=42, entity="thing")),
    ),
)
def test_forbid_kwarg_consts_triggered(kls, kwargs):
    with pytest.raises(TypeError):
        type("TestError",
             (kls,),
             {"__features__": root.Features.FORBID_KWARG_CONSTS}
             )(**kwargs)


@pytest.mark.parametrize("features", tuple(root.Features(i) for i in range(7)))
@mock.patch("izulu._utils.check_kwarg_consts")
@mock.patch("izulu._utils.check_undeclared_fields")
@mock.patch("izulu._utils.check_missing_fields")
def test_process_features(mock_missing, mock_undeclared, mock_const, features):
    with mock.patch.object(errors.RootError,
                           "__features__",
                           new_callable=mock.PropertyMock) as mocked:
        mocked.return_value = features

        e = errors.RootError()
        args = (e._Error__cls_store, frozenset(e._Error__kwargs))

        if root.Features.FORBID_MISSING_FIELDS in features:
            mock_missing.assert_called_once_with(*args)
        else:
            mock_missing.assert_not_called()
        if root.Features.FORBID_UNDECLARED_FIELDS in features:
            mock_undeclared.assert_called_once_with(*args)
        else:
            mock_undeclared.assert_not_called()
        if root.Features.FORBID_KWARG_CONSTS in features:
            mock_const.assert_called_once_with(*args)
        else:
            mock_const.assert_not_called()


def test_feature_presets():
    default = (
            root.Features.FORBID_MISSING_FIELDS
            | root.Features.FORBID_UNDECLARED_FIELDS
            | root.Features.FORBID_KWARG_CONSTS
            | root.Features.FORBID_NON_NAMED_FIELDS
    )

    assert root.Features.NONE is root.Features(0)
    assert root.Features.NONE == root.Features(0)
    assert root.Features.DEFAULT is default
    assert root.Features.DEFAULT == default


def test_default_features():
    assert root.Error.__features__ is root.Features.DEFAULT



import datetime
import types
from unittest import mock

import pytest

from izulu import _utils
from izulu import root
from tests import errors


TS = datetime.datetime.now()


@mock.patch("izulu.root.Error._hook")
@mock.patch("izulu.root.Error._Error__process_template")
@mock.patch("izulu.root.Error._Error__populate_attrs")
@mock.patch("izulu.root.Error._Error__process_features")
def test_init(fake_proc_ftrs, fake_set_attrs, fake_proc_tpl, fake_hook):
    fake_proc_tpl.return_value = errors.RootError.__template__
    overriden_message = "overriden message"
    fake_hook.return_value = overriden_message
    store = getattr(errors.RootError, "_Error__cls_store")
    manager = mock.Mock()
    manager.attach_mock(fake_proc_ftrs, "fake_proc_ftrs")
    manager.attach_mock(fake_set_attrs, "fake_set_attrs")
    manager.attach_mock(fake_proc_tpl, "fake_proc_tpl")
    manager.attach_mock(fake_hook, "fake_hook")
    fake_hook_call = mock.call.fake_hook(store,
                                         {},
                                         errors.RootError.__template__)
    expected_calls = [mock.call.fake_proc_ftrs(),
                      mock.call.fake_set_attrs(),
                      mock.call.fake_proc_tpl({}),
                      fake_hook_call]

    e = errors.RootError()

    assert manager.mock_calls == expected_calls
    assert str(e) == overriden_message


@pytest.mark.parametrize(
    ("kls", "kwargs", "msg", "attrs", "not_attrs"),
    (
            (errors.RootError, dict(), "Unspecified error", dict(), tuple()),
            (errors.TemplateOnlyError,
             dict(name="John", age=42),
             "The John is 42 years old",
             dict(),
             ("name", "age")),
            (errors.ComplexTemplateOnlyError,
             dict(name="John", age=42, ts=TS),
             ("********John********  42.000000"
              f" 0b101010 {TS:%Y-%m-%d %H:%M:%S}"),
             dict(),
             ("name", "age", "ts")),
            (errors.AttributesOnlyError,
             dict(name="John", age=42),
             "Static message template",
             dict(name="John", age=42),
             tuple()),
            (errors.AttributesWithStaticDefaultsError,
             dict(name="John"),
             "Static message template",
             dict(name="John", age=0),
             tuple()),
            (errors.AttributesWithDynamicDefaultsError,
             dict(name="John"),
             "Static message template",
             dict(name="John", age=0),
             tuple()),
            (errors.ClassVarsError,
             dict(),
             "Static message template",
             dict(name="Username", age=42),
             tuple()),
            (errors.MixedError,
             dict(name="John", age=10, note="...", timestamp=TS),
             "The John is 10 years old with ...",
             dict(entity="The Entity", name="John", age=10, timestamp=TS,
                  my_type="MixedError"),
             ("note",)),
            (errors.DerivedError,
             dict(name="John", age=10, note="...", timestamp=TS,
                  surname="Brown", box=dict(a=11)),
             "The John Brown is 10 years old with ...",
             dict(entity="The Entity", name="John", age=10, timestamp=TS,
                  my_type="DerivedError", surname="Brown", box=dict(a=11)),
             ("note",)),
    )
)
def test_instantiate_ok(kls, kwargs, msg, attrs, not_attrs):
    e = kls(**kwargs)

    assert str(e) == msg
    for attr, value in attrs.items():
        assert getattr(e, attr) == value
    for attr in not_attrs:
        assert not hasattr(e, attr)


@pytest.mark.parametrize(
    ("kls", "kwargs"),
    (
            (errors.RootError, dict(name="John")),
            (errors.TemplateOnlyError, dict(name="John")),
            (errors.AttributesOnlyError, dict(age=42)),
            (errors.AttributesWithStaticDefaultsError, dict()),
            (errors.AttributesWithDynamicDefaultsError, dict(age=0)),
            (errors.ClassVarsError, dict(name="John")),
    )
)
def test_instantiate_fail(kls, kwargs):
    with pytest.raises(TypeError):
        kls(**kwargs)


@pytest.mark.parametrize(
    ("kls", "kwargs", "expected_kwargs"),
    (
            (errors.RootError, dict(), dict()),
            (errors.RootError, dict(name="John"), dict(name="John")),
            (errors.ClassVarsError, dict(), dict(name="Username", age=42)),
            (errors.ClassVarsError,
             dict(name="John"),
             dict(name="John", age=42)),
            (errors.MixedError,
             dict(name="John", surname="Brown", updated_at=TS, timestamp=TS),
             dict(name="John", surname="Brown", age=0,
                  my_type="MixedError", entity="The Entity",
                  updated_at=TS, timestamp=TS)),
            (errors.MixedError,
             dict(updated_at=TS, timestamp=TS),
             dict(age=0, my_type="MixedError", entity="The Entity",
                  updated_at=TS, timestamp=TS)),
            (errors.DerivedError,
             dict(name="John", surname="Brown", updated_at=TS, timestamp=TS),
             dict(name="John", surname="Brown", full_name="John Brown", age=0,
                  my_type="DerivedError", entity="The Entity",
                  location=(50.3, 3.608), updated_at=TS, timestamp=TS)),
    )
)
@mock.patch("izulu._utils.format_template")
def test_process_template(mock_format, kls, kwargs, expected_kwargs):
    with mock.patch.object(kls,
                           "__features__",
                           new_callable=mock.PropertyMock) as mocked:
        mocked.return_value = root.Features.NONE
        kls(**kwargs)

    mock_format.assert_called_once_with(kls.__template__, expected_kwargs)


def test_hook():
    e = errors.RootError()
    orig_msg = "my message"

    final_msg = e._hook(e._Error__cls_store, dict(), orig_msg)

    assert final_msg == orig_msg
    assert id(final_msg) == id(orig_msg)


def test_constants():
    e = errors.RootError()

    assert e.__template__ == "Unspecified error"
    assert e.__features__ == root.Features.DEFAULT
    assert e._Error__cls_store == _utils.Store(
        fields=frozenset(),
        const_hints=types.MappingProxyType(dict()),
        inst_hints=types.MappingProxyType(dict()),
        consts=types.MappingProxyType(dict()),
        defaults=frozenset(),
    )



import datetime
from unittest import mock

import pytest

from tests import errors


TS = datetime.datetime.now()


@pytest.mark.parametrize(
    "kwargs",
    (
            dict(),
            dict(name="John"),
            dict(age=42),
            dict(name="John", age=42),
            dict(name="John", age=42, ts=53452345.3465),
            dict(name="John", age="Karl", ts=TS),
    ),
)
def test_templating(kwargs):
    with pytest.raises(ValueError):
        errors.ComplexTemplateOnlyError(**kwargs)


@pytest.mark.parametrize(
    ("err", "expected"),
    (
            (errors.RootError(), "Unspecified error"),
            (errors.MixedError(name="John", age=10, note="..."),
             "The John is 10 years old with ..."),
    )
)
def test_str(err, expected):
    assert str(err) == expected


@pytest.mark.parametrize(
    ("err", "expected"),
    (
            (errors.RootError(), "izulu.root.Error()"),
            (errors.MixedError(name="John", age=10, note="...", timestamp=TS),
             ("tests.errors.MixedError(name='John', age=10,"
              f" note='...', timestamp={TS!r}, my_type='MixedError')")),
    )
)
def test_repr(err, expected):
    assert repr(err) == expected


def test_repl_repr():
    e = errors.TemplateOnlyError(name="John", age=42)

    with mock.patch.object(errors.TemplateOnlyError,
                           "__module__",
                           new_callable=mock.PropertyMock) as mocked:
        mocked.return_value = "__main__"

        assert repr(e) == "__main__.TemplateOnlyError(name='John', age=42)"


@pytest.mark.parametrize(
    ("err", "expected"),
    (
            (errors.RootError(), "Error: Unspecified error"),
            (errors.MixedError(name="John", age=10, note="..."),
             "MixedError: The John is 10 years old with ..."),
    )
)
def test_as_str(err, expected):
    assert err.as_str() == expected



import types

from izulu import _utils


def _make_store_kwargs(fields=None,
                       const_hints=None, inst_hints=None,
                       consts=None, defaults=None) -> dict:
    return dict(
        fields=frozenset(fields or tuple()),
        const_hints=types.MappingProxyType(const_hints or dict()),
        inst_hints=types.MappingProxyType(inst_hints or dict()),
        consts=types.MappingProxyType(consts or dict()),
        defaults=frozenset(defaults or tuple()),
    )


def _make_store(fields=None,
                const_hints=None, inst_hints=None,
                consts=None, defaults=None) -> _utils.Store:
    return _utils.Store(
        **_make_store_kwargs(fields=fields,
                             inst_hints=inst_hints,
                             const_hints=const_hints,
                             consts=consts,
                             defaults=defaults)
    )



import datetime
import types
from unittest import mock

import pytest

from izulu import root
from tests import errors


@pytest.mark.parametrize(
    ("kls", "fields", "hints", "registered", "defaults", "consts"),
    (
            (
                    errors.RootError,
                    frozenset(),
                    types.MappingProxyType({}),
                    frozenset(),
                    frozenset(),
                    types.MappingProxyType({}),
            ),
            (
                    errors.TemplateOnlyError,
                    frozenset(("name", "age")),
                    types.MappingProxyType({}),
                    frozenset(("name", "age")),
                    frozenset(),
                    types.MappingProxyType({}),
            ),
            (
                    errors.AttributesOnlyError,
                    frozenset(),
                    types.MappingProxyType(dict(name=str, age=int)),
                    frozenset(("name", "age")),
                    frozenset(),
                    types.MappingProxyType({}),
            ),
            (
                    errors.AttributesWithStaticDefaultsError,
                    frozenset(),
                    types.MappingProxyType(dict(name=str, age=int)),
                    frozenset(("name", "age")),
                    frozenset(("age",)),
                    types.MappingProxyType({}),
            ),
            (
                    errors.AttributesWithDynamicDefaultsError,
                    frozenset(),
                    types.MappingProxyType(dict(name=str, age=int)),
                    frozenset(("name", "age")),
                    frozenset(("age",)),
                    types.MappingProxyType({}),
            ),
            (
                    errors.ClassVarsError,
                    frozenset(),
                    types.MappingProxyType({}),
                    frozenset(),
                    frozenset(),
                    types.MappingProxyType(dict(name="Username", age=42)),
            ),
            (
                    errors.MixedError,
                    frozenset(("name", "age", "note")),
                    types.MappingProxyType(dict(name=str,
                                                age=int,
                                                timestamp=datetime.datetime,
                                                my_type=str)),
                    frozenset(("name", "age", "note", "timestamp", "my_type")),
                    frozenset(("age", "timestamp", "my_type")),
                    types.MappingProxyType(dict(entity="The Entity")),
            ),
            (
                    errors.DerivedError,
                    frozenset(("name", "surname", "age", "note")),
                    types.MappingProxyType(dict(name=str,
                                                age=int,
                                                timestamp=datetime.datetime,
                                                my_type=str,
                                                surname=str,
                                                location=tuple[float, float],
                                                updated_at=datetime.datetime,
                                                full_name=str,
                                                box=dict)),
                    frozenset(("name",
                               "age",
                               "note",
                               "timestamp",
                               "my_type",
                               "surname",
                               "location",
                               "updated_at",
                               "full_name",
                               "box")),
                    frozenset(("age",
                               "timestamp",
                               "my_type",
                               "location",
                               "updated_at",
                               "full_name")),
                    types.MappingProxyType(dict(entity="The Entity")),
            ),
    )
)
def test_cls_store(kls, fields, hints, registered, defaults, consts):
    """Validates store management from root.Error.__init_subclass__"""

    store = getattr(kls, "_Error__cls_store")

    assert type(store.fields) is type(fields)
    assert store.fields == fields
    assert type(store.inst_hints) is type(hints)
    assert store.inst_hints == hints
    assert type(store.registered) is type(registered)
    assert store.registered == registered
    assert type(store.defaults) is type(defaults)
    assert store.defaults == defaults
    assert type(store.consts) is type(consts)
    assert store.consts == consts


@pytest.mark.parametrize(
    "features",
    (root.Features.FORBID_NON_NAMED_FIELDS,
     root.Features.NONE)
)
@mock.patch("izulu._utils.check_non_named_fields", return_value=0)
def test_cls_validation(mocked_check, features):
    """Validates feature checks from root.Error.__init_subclass__"""

    type("Err", (errors.ClassVarsError,), {"__features__": features})

    if features is root.Features.NONE:
        mocked_check.assert_not_called()
    else:
        mocked_check.assert_called_once()
